#!/usr/bin/python
#
# Copyright (c) Andrea Micheloni 2011
#
# mDNS/DNS-SD tracker for the lazy social engineer
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, avahi, dbus, datastore_interface

try:
    import dbus.glib
except ImportError:
    pass

class AvahiBrowser:
	def __init__(self):
		self.bus = dbus.SystemBus()
		self.avahiServer = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER), avahi.DBUS_INTERFACE_SERVER)
		self.avahiServerVersion = self.avahiServer.GetVersionString()
		self.dataStore = datastore_interface.DataStore()
		self.browseServiceTypes()

	def browseServiceTypes(self):
		self.serviceTypes = {}
		print "Browsing service types...";

		serviceTypeBrowser = self.avahiServer.ServiceTypeBrowserNew(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC, "local", dbus.UInt32(0))

		browser = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME, serviceTypeBrowser), avahi.DBUS_INTERFACE_SERVICE_TYPE_BROWSER)
		browser.connect_to_signal('ItemNew', self.foundServiceType)
		browser.connect_to_signal('ItemRemove', self.removeServiceType)
		browser.connect_to_signal('AllForNow', self.endBrowsing)

	def foundServiceType(self, interface, protocol, stype, domain, flags):
		print "+T", stype, "\t"
		if not stype in self.serviceTypes:
			self.serviceTypes[stype] = ServiceTypeBrowser(self, interface, protocol, stype, domain, flags)

	def removeServiceType(self, interface, protocol, stype, domain, flags):
		print "-T", stype, "\t"

	def endBrowsing(self):
		print "=A All for now"

	def __repr__(self):
		return "AvahiBrowser("+str(self.serviceTypes)+")"

class ServiceTypeBrowser:
	def __init__(self, avahiBrowser, interface, protocol, stype, domain, flags):
		self.avahiBrowser = avahiBrowser
		self.interface = interface
		self.protocol = protocol
		self.stype = stype
		self.domain = domain
		self.flags = flags
		self.key = self.avahiBrowser.dataStore.storeType(str(stype))

		self.browseServices()

	def browseServices(self):
		self.services = {}
		serviceBrowser = self.avahiBrowser.avahiServer.ServiceBrowserNew(self.interface, self.protocol, self.stype, self.domain, dbus.UInt32(0))

		browser = dbus.Interface(self.avahiBrowser.bus.get_object(avahi.DBUS_NAME, serviceBrowser), avahi.DBUS_INTERFACE_SERVICE_BROWSER)
		browser.connect_to_signal('ItemNew', self.foundService)
		browser.connect_to_signal('ItemRemove', self.removeService)

	def foundService(self, interface, protocol, name, stype, domain, flags):
		print "+S", stype, "\t", "\"" + unicode(name) + "\""
		if not (interface, domain, name) in self.services:
			self.services[(interface, domain, name)] = ServiceHandler(self.avahiBrowser, self, unicode(name))
		self.services[(interface, domain, name)].saveAppeared()
		self.services[(interface, domain, name)].resolveService()

	def removeService(self, interface, protocol, name, stype, domain, flags):
		print "-S", stype, "\t", "\"" + unicode(name) + "\""
		self.services[(interface, domain, name)].saveDisappeared()

	def __repr__(self):
		return "ServiceTypeBrowser(" + str(self.avahiBrowser) + ", " + str(self.interface) + ", " + str(self.protocol) + ", " + str(self.stype) + ", " + str(self.domain) + ", " +  str(self.flags) + ")"

class ServiceHandler:
	def __init__(self, avahiBrowser, serviceType, name):
		self.avahiBrowser = avahiBrowser
		self.serviceType = serviceType
		self.name = name
		self.key = self.avahiBrowser.dataStore.storeService(self.serviceType.key, name)
		self.serviceResolver = None

	def saveAppeared(self):
		self.insertionKey = self.avahiBrowser.dataStore.storeServiceAppearance(self.key, True)

	def saveDisappeared(self):
		self.avahiBrowser.dataStore.storeServiceAppearance(self.key, False)

	def resolveService(self):
		if self.serviceResolver == None:
			self.serviceResolver = ServiceResolver(self.avahiBrowser, self.serviceType, self)
		self.serviceResolver.askResolution()

	def __repr__(self):
		return "ServiceHandler(" + str(self.avahiBrowser) + ", " + str(self.serviceType) + ", " + unicode(self.name) + ")"

class ServiceResolver:
	def __init__(self, avahiBrowser, serviceType, serviceHandler):
		self.avahiBrowser = avahiBrowser
		self.serviceType = serviceType
		self.serviceHandler = serviceHandler

	def askResolution(self):
		self.avahiBrowser.avahiServer.ResolveService(self.serviceType.interface, self.serviceType.protocol, unicode(self.serviceHandler.name), str(self.serviceType.stype), str(self.serviceType.domain), avahi.PROTO_UNSPEC, dbus.UInt32(0),	reply_handler=self.resolutionResult, error_handler=self.resolutionError)

	def resolutionError(self, err):
		print "-R", str(err)

	def resolutionResult(self, interface, protocol, name, stype, domain, host, proto, address, port, txt, flags):
		print "=R", stype, "\t", "\"" + unicode(name) + "\"", str(address)+":"+str(port), str(avahi.txt_array_to_string_array(txt))
		self.avahiBrowser.dataStore.storeServiceResolution(self.serviceHandler.insertionKey, host, address, port, str(avahi.txt_array_to_string_array(txt)))

